/* 
 * The file was modified by Sunner Sun on 31-10-04 23:50
 * to fit the courses in HIT.
 */
  
/*
 * File: simpio.c
 * Version: 3.0
 * Last modified on Tue Oct  4 11:24:40 1994 by eroberts
 * -----------------------------------------------------
 * This file implements the simpio.h interface.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <assert.h>

#include "simpio.h"

/*
 * Constants:
 * ----------
 * InitialBufferSize -- Initial buffer size for ReadLine
 * ErrorExitStatus -- Status value used in exit call
 */

#define InitialBufferSize 120
#define ErrorExitStatus 1

/*
 * Function: GetBlock
 * Usage: ptr = (type) GetBlock(nbytes);
 * -------------------------------------
 * GetBlock allocates a block of memory of the given size.  If
 * no memory is available, GetBlock generates an error.
 */

static void *GetBlock(size_t nbytes);

/*
 * Function: FreeBlock
 * Usage: FreeBlock(ptr);
 * ----------------------
 * FreeBlock frees the memory associated with ptr, which must
 * have been allocated using GetBlock.
 */

static void FreeBlock(void *ptr);

/*
 * Function: GetValue
 * Usage: GetValue(format, &value, errorMessage);
 * ----------------------
 * GetValue first read a line and then call sscanf to translate 
 * the number to value whose type matches format. Reading an entire 
 * line is essential to good error recovery, because the characters 
 * after the point of error would otherwise remain in the input buffer
 * and confuse subsequent input operations. The sscanf line allows 
 * white space before and after the number but no other extraneous 
 * characters. If the input doesn't match format, errorMessage will be
 * printed out.
 */

static void GetValue(const char* format, void* pValue, const char* errorMessage);

/*
 * Function: GetLine
 * Usage: s = GetLine();
 * ---------------------
 * GetLine reads a line of text from standard input and returns
 * the line as a string.  The newline character that terminates
 * the input is not stored as part of the string.
 */

static char* GetLine(void);

/* Exported entries */

/*
 * Functions: GetInt, GetLong, GetReal, GetChar 
 * ---------------------------------------
 * These functions call GetValue.
 */

int GetInt(void)
{
    int value;
    GetValue("%d", &value, "Please input an integer number.\n");
    return value;
}

long GetLong(void)
{
    return (int)GetInt();
}

double GetReal(void)
{
    double value;
    GetValue("%lf", &value, "Please input a real number.\n");
    return value;
}

char GetChar(void)
{
    char ch;
    GetValue("%c", &ch, "Please input one character.\n");
    return ch; 
}

/* static entries */

static void GetValue(const char* format, void* pValue, const char* errorMessage)
{
    char* line;
    char termch;
    char template[12];
    int ret;
    
    assert(strlen(format) <= 5);

    do
    {
        line = GetLine();
        if (line == NULL) 
            Error("GetValue: unexpected end of file");

        sprintf(template, " %s %%c", format);
        assert(strlen(template) < 12);
        ret = sscanf(line, template, pValue, &termch);
        
        if (ret == 1)
        {
            break;
        }
        else
        {
            if (ret == 2)
            {
                printf("Unexpected character: '%c'\n", termch);
            }
            printf(errorMessage);
        }
        printf("Retry: ");
    } while (1);

    FreeBlock(line);
}

/*
 * Function: GetLine
 * -----------------
 * This function operates by reading characters from the stdin 
 * into a dynamically allocated buffer.  If the buffer becomes
 * full before the end of the line is reached, a new buffer 
 * twice the size of the previous one is allocated.
 */

static char* GetLine(void)
{
    char *line, *nline;
    int n, ch, size;

    n = 0;
    size = InitialBufferSize;
    line = (char*)GetBlock(size + 1);
    while ((ch = getchar()) != '\n' && ch != EOF) 
    {
        if (n == size) 
        {
            /* double the buffer */
            size *= 2;
            nline = (char*) GetBlock(size + 1);
            strncpy(nline, line, n);
            FreeBlock(line);
            line = nline;
        }
        line[n++] = ch;
    }

    if (n == 0 && ch == EOF) 
    {
        FreeBlock(line);
        return (NULL);
    }
    else
    {
        line[n] = '\0';
        return (line);
    }    
}

static void *GetBlock(size_t nbytes)
{
    void *result;

    result = malloc(nbytes);
    if (result == NULL) 
        Error("No memory available");
    return (result);
}

static void FreeBlock(void *ptr)
{
    free(ptr);
}

void Error(char* msg, ...)
{
    va_list args;

    va_start(args, msg);
    fprintf(stderr, "Error: ");
    vfprintf(stderr, msg, args);
    fprintf(stderr, "\n");
    va_end(args);
    exit(ErrorExitStatus);
}

